import { MessageTransmissionTransmitter } from "./msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "./msg.transmission.receiver";
import { ConnectionManager } from "../connector/connector.manager";
import { EventMessage, MessageTransmission, MessageTransmissionManager as MessageTransmissionManagerInterface, ReceiverProfile, TransmitterProfile } from "../interface/transport.interface";
import { v4 as uuidv4 } from 'uuid'
import { AdapterSet, TransportEvent, Event } from "../interface/connector.interface";
import { MessageTransmissionRequestResponse } from "./msg.transmission.request-response";
import { filter, Observable, Observer, Subject } from "rxjs";

export class MessageTransmissionManager implements MessageTransmissionManagerInterface {
    private browserEnv!: boolean
    transmission: MessageTransmission[] = []
    connectionManager!: ConnectionManager
    event!: Subject<TransportEvent>

    constructor(event: Subject<TransportEvent>, browserEnv?: boolean) {
        if (browserEnv) this.browserEnv = browserEnv
        // logic here
        console.log(`TransmissionManager: Contructing Transmission Manager...`)
        this.event = event
        this.connectionManager = new ConnectionManager(this.event, browserEnv)

        // this.event.subscribe(event => console.log(`event`, event))

        // note that if this server is down, all these instances of transmission and connector would be lost as well. SO cannot just simply find "instances" and reuse them. Must reinstantiate them again
        this.handleEvent(`Client Disconnected` as Event, event)
        this.handleEvent('Client Reconnected' as Event, event)
    }

    /* so there will be some changes here. will nto be assigning just one, but all of them dynamically to pour into this boy
    Transmitter only have to call this once. */
    subscribe(): Observable<MessageTransmission> {
        return new Observable((observer: Observer<MessageTransmission>) => {
            const targetEvent: Event = this.browserEnv ? 'New Server' : 'New Client';
            this.event.pipe(
                filter(event => event.event == targetEvent)
            ).subscribe(event => {
                // get all adapters for all the connection
                observer.next(this.instantiateComponents((event.data as EventMessage).clientId))
            })
        })
    }

    private instantiateComponents(clientId: string): MessageTransmission {
        console.log(`Instantiating new transmission set for another ${this.browserEnv ? 'Server' : 'Client'}`)
        let adapterSet: AdapterSet[] = []
        if (this.connectionManager.getTransportArray().length > 0) {
            this.connectionManager.getTransportArray().forEach(transport => {
                adapterSet.push(this.connectionManager.getAdapter(clientId, transport))
            })
        } else {
            throw new Error(`Transmission Manager: No transport is Instantiated`)
        }

        // 1 set only
        let transmitter: MessageTransmissionTransmitter = this.getTransmitter(clientId, adapterSet, this.event.asObservable())
        let receiver: MessageTransmissionReceiver = this.getReceiver(clientId, adapterSet, this.event.asObservable())
        let requestResponse: MessageTransmissionRequestResponse = this.getRequestResponse(transmitter, receiver, this.event.asObservable())
        let transmission: MessageTransmission = {
            id: clientId,
            transmitter: transmitter,
            receiver: receiver,
            requestResponse: requestResponse,
            event: this.event.asObservable()
        }
        this.transmission.push(transmission)
        return transmission
    }


    private getTransmitter(transmissionId: string, adapterSets: AdapterSet[], event: Observable<TransportEvent>): MessageTransmissionTransmitter {
        let transmitterProfile: TransmitterProfile = {
            id: transmissionId,
            name: '', // for now make it empty. We will use the assigned uuid here
            dateCreated: new Date()
        }
        return new MessageTransmissionTransmitter(transmitterProfile, adapterSets, event)
    }

    private getReceiver(transmissionId: string, adapterSets: AdapterSet[], event: Observable<TransportEvent>): MessageTransmissionReceiver {
        let receiverProfile: ReceiverProfile = {
            id: transmissionId,
            name: '', // for now make it empty. We will use the assigned uuid here
            dateCreated: new Date()
        }
        return new MessageTransmissionReceiver(receiverProfile, adapterSets, event)
    }

    private getRequestResponse(transmitterInstance: MessageTransmissionTransmitter, receiverInstance: MessageTransmissionReceiver, event: Observable<TransportEvent>): MessageTransmissionRequestResponse {
        return new MessageTransmissionRequestResponse(transmitterInstance, receiverInstance, event)
    }



    private handleEvent(eventName: Event, eventObs: Observable<TransportEvent>): void {
        eventObs.pipe(
            filter((event: TransportEvent) => event.event === eventName)
        ).subscribe(event => {
            // assuming this is reconnection case
            if (event.event == 'Client Reconnected') {
                this.reconnectionHandler((event.data as EventMessage).clientId)
            }

            // can include more event handlers here
        })
    }

    private reconnectionHandler(clientId: string): void {
        let transmissionObj: MessageTransmission | undefined = Array.from(this.transmission).find(obj => obj.id === clientId)
        if (!transmissionObj) {
            let transmission: MessageTransmission = this.instantiateComponents(clientId)
            this.transmission.push(transmission)
        }
    }

}


