import { MessageTransmissionBase } from "./msg.transmission.base";
import { EventMessage, FisMessage, MessageTransmitter as MessageTransmitterInterface, TransmitterProfile } from '../interface/transport.interface'
import { AdapterSet, ConnectionState, Event, Transport, TransportEvent, TransportMessage } from "../interface/connector.interface";
import { v4 as uuidv4 } from 'uuid'
import { TransmitterConnectionAdapter } from "../connector/connector.transmitter";
import { BehaviorSubject, filter, map, Observable, Subject } from "rxjs";
import { RetransmissionService } from "../utils/retransmission.service";
import { WrappedMessage } from "../utils/message.ordering";

/* Take in all the messages that needs to be transported, and divide them accordingly. So the connector instances will do just that
connectors or adapters will have their own identifier*/
export class MessageTransmissionTransmitter extends MessageTransmissionBase implements MessageTransmitterInterface {
    private messageToBeTransmitted!: Subject<FisMessage>
    transmitterProfile!: TransmitterProfile;
    retransmission!: RetransmissionService;

    constructor(profile: TransmitterProfile, adapterSets: AdapterSet[], event: Observable<TransportEvent>) {
        super(event)
        this.event = event
        this.messageToBeTransmitted = new Subject()
        this.retransmission = new RetransmissionService()
        this.setTransmitter(profile)
        this.setUpAdapter(adapterSets)
        this.setUpRetransmission()
    }
    
    // by the time this transmission set is instantiated, the connected client would've been online. Need ot manually signal retransmission to release buffer immediately
    setUpRetransmission(): void {
        let connectionStateEvent = new BehaviorSubject<'OFFLINE' | 'ONLINE'>('ONLINE')
        this.event.pipe(
            filter(event => event.event == 'Client Disconnected' || event.event == 'Client Reconnected'),
            filter(event => (event.data as EventMessage).clientId == this.transmitterProfile.id),
            map(event => {
                if (event.event == 'Client Disconnected') {
                    return 'OFFLINE'
                } else {
                    return `ONLINE`
                }
            })
        ).subscribe((signal: ConnectionState) => {
            connectionStateEvent.next(signal)
        })
        this.retransmission.implementRetransmission(this.messageToBeTransmitted, connectionStateEvent, true)
        // automatically subscribe to allow released bffered messages to be released
        this.retransmission.returnSubjectForBufferedItems().subscribe((message: WrappedMessage) => {
            // need to work with wrapped messages
            (this.adapterService as TransmitterConnectionAdapter).emit(message)
        })
    }
    
    setTransmitter(transmitterProfile: TransmitterProfile): void {
        this.transmitterProfile = transmitterProfile
    }

    emit(message: FisMessage): void {
        // for now only use one adapter
        this.messageToBeTransmitted.next(message)
    }

    setUpAdapter(adapterSets: AdapterSet[]) {
        if (adapterSets.length > 0) {
            adapterSets.forEach((adapter: AdapterSet) => {
                this.adaptorsArray.push(adapter.transmitterAdapter)
            })
        }
        // for now just hardcode to use 1 adapter type until connection manager is further enhacne to configure adapters on the fly
        this.adapterService = this.adaptorsArray.find(obj => obj.getInfo().transportType === Transport.Websocket)
    }

}